// ignore_for_file: close_sinks

import 'dart:async';
import 'dart:convert';

import 'package:another_flushbar/flushbar.dart';
import 'package:flutter/cupertino.dart';
import 'package:http/http.dart';
import 'package:pam_flutter/repositories/types.dart';
import 'package:pam_flutter/utils/requester.dart';
import 'package:pam_flutter/utils/time_helper.dart';
import 'package:rxdart/rxdart.dart';

class BaseDataRepository<T> implements IBaseDataRepository {
  final BuildContext buildCtx;
  final IRepositoryOptions options;

  bool _isLoading = false;
  bool _isLoaded = false;
  bool _isError = false;
  String _errorMessage = "";
  List<dynamic>? _items;
  dynamic _data;

  late StreamController<bool> _isLoadingSC;
  late StreamController<bool> _isLoadedSC;
  late StreamController<bool> _isErrorSC;
  late StreamController<String> _errorMessageSC;
  late StreamController<List<T>?> _itemsSC;
  late StreamController<T?> _dataSC;

  BaseDataRepository(
    this.buildCtx,
    this.options,
  ) {
    this.initial();
  }

  @override
  bool get isLoading => this._isLoading;

  @override
  bool get isLoaded => this._isLoaded;

  @override
  bool get isError => this._isError;

  @override
  String get errorMessage => this._errorMessage;

  @override
  List<T> get items => this.transforms(this._items);

  @override
  T? get data => this.transform(this._data);

  @override
  StreamController<bool> get isLoadingSC => this._isLoadingSC;

  @override
  StreamController<bool> get isLoadedSC => this._isLoadedSC;

  @override
  StreamController<bool> get isErrorSC => this._isErrorSC;

  @override
  StreamController<String> get errorMessageSC => this._errorMessageSC;

  @override
  StreamController<List<T>?> get itemsSC => this._itemsSC;

  @override
  StreamController<T?> get dataSC => this._dataSC;

  @override
  void toLoadingStatus() {
    this._isLoading = true;
    this._isLoadingSC.add(true);

    this._isLoaded = false;
    this._isLoadedSC.add(false);

    this._isError = false;
    this._isErrorSC.add(false);
  }

  @override
  void toLoadedStatus() {
    this._isLoading = false;
    this._isLoadingSC.add(false);

    this._isLoaded = true;
    this._isLoadedSC.add(true);
  }

  @override
  void toErrorStatus(e) {
    this._isLoading = false;
    this._isLoadingSC.add(false);

    this._isError = true;
    this._isErrorSC.add(true);

    this._errorMessage = e.toString();
    this._errorMessageSC.add(e.toString());

    print("error: ${e.toString()}");
  }

  @override
  List<T> transforms(tss) {
    return List<T>.empty(growable: true);
  }

  @override
  T? transform(ts) {
    return null;
  }

  @override
  Future<void> fetch({Map<String, dynamic>? params, bool isMock: false}) async {
    try {
      this.toLoadingStatus();
      late Map<String, dynamic> data;

      if (params == null) {
        params = {};
      }

      if (isMock) {
        await TimeHelper.sleep();
        data = {"items": this.options.getMockItems()};
      } else {
        Response response =
            await Requester.get(this.options.getBaseUrl(), params);
        Map<String, dynamic> js = json.decode(utf8.decode(response.bodyBytes));
        data = js;
      }

      this._items = data["items"];
      this.itemsSC.add(this.items);

      this.toLoadedStatus();
    } catch (e) {
      this.alertError(e);
      this.toErrorStatus(e);
    }
  }

  @override
  Future<void> fetchAfterId(String afterId,
      {Map<String, dynamic>? params, bool isMock = false}) async {
    try {
      this.toLoadingStatus();
      late Map<String, dynamic> data;

      if (params == null) {
        params = {};
      }

      if (isMock) {
        await TimeHelper.sleep();
        data = {"items": this.options.getMockItems()?.sublist(0, 6)};
      } else {
        Response response =
            await Requester.get(this.options.getBaseUrl(), params);
        Map<String, dynamic> js = json.decode(utf8.decode(response.bodyBytes));
        data = js;
      }

      this._items = [
        ...(this._items ?? []),
        ...(data["items"] ?? []),
      ];
      this.itemsSC.add(this.items);

      this.toLoadedStatus();
    } catch (e) {
      this.alertError(e);
      this.toErrorStatus(e);
    }
  }

  @override
  Future<void> get(String id,
      {Map<String, dynamic>? params, bool isMock: false}) async {
    try {
      this.toLoadingStatus();
      late Map<String, dynamic> data;

      if (params == null) {
        params = {};
      }

      if (isMock) {
        await TimeHelper.sleep();
        data = {"data": this.options.getMockItem()};
      } else {
        Response response = await Requester.get(
            "${this.options.getFindUrl() ?? this.options.getBaseUrl()}"
                .replaceFirst(":id", id),
            params);
        Map<String, dynamic> js = json.decode(utf8.decode(response.bodyBytes));
        data = {"data": js};
      }

      this._data = data["data"];
      this.dataSC.add(this.data);

      this.toLoadedStatus();
    } catch (e) {
      this.alertError(e);
      this.toErrorStatus(e);
    }
  }

  @override
  void forceValueNotify() {
    this._isLoadingSC.add(true);
    this._isLoadedSC.add(false);
    this._isErrorSC.add(false);
    this._errorMessageSC.add("");

    this._itemsSC.add(null);
    this._dataSC.add(null);

    this._isLoadingSC.add(this.isLoading);
    this._isLoadedSC.add(this.isLoaded);
    this._isErrorSC.add(this.isError);
    this._errorMessageSC.add(this.errorMessage);

    this._itemsSC.add(this.items);
    this._dataSC.add(this.data);
  }

  @override
  void setInnerData(data) {
    return this._data = data;
  }

  @override
  void initial() {
    this._isLoadingSC = BehaviorSubject<bool>();
    this._isLoadedSC = BehaviorSubject<bool>();
    this._isErrorSC = BehaviorSubject<bool>();
    this._errorMessageSC = BehaviorSubject<String>();

    this._itemsSC = BehaviorSubject<List<T>?>();
    this._dataSC = BehaviorSubject<T?>();
  }

  @override
  void clear() {
    this._isLoading = false;
    this._isLoaded = false;
    this._isError = false;
    this._errorMessage = "";
    this._isLoadingSC = BehaviorSubject<bool>();
    this._isLoadedSC = BehaviorSubject<bool>();
    this._isErrorSC = BehaviorSubject<bool>();
    this._errorMessageSC = BehaviorSubject<String>();

    this._itemsSC = BehaviorSubject<List<T>?>();
    this._dataSC = BehaviorSubject<T?>();

    this._items = null;
    this._data = null;
  }

  @override
  void dispose() {
    this._isLoadingSC.close();
    this._isLoadedSC.close();
    this._isErrorSC.close();
    this._errorMessageSC.close();
  }

  @override
  void alertError(dynamic e) {
    String title = "Error";
    dynamic message = e;

    print("error = ${e.toString()}");

    if (e is Map && message is Map) {
      title = e["code"] ?? "Error";
      message = e["message"] ?? e;
    }

    Flushbar(
      title: title,
      message: message.toString(),
      duration: Duration(seconds: 3),
      flushbarStyle: FlushbarStyle.GROUNDED,
    )..show(this.buildCtx);
  }
}

class NewRepositoryOptions implements IRepositoryOptions {
  final String baseUrl;
  final String? findUrl;
  final String? addUrl;
  final String? updateUrl;
  final String? deleteUrl;
  final List<Map<String, dynamic>>? mockItems;
  final Map<String, dynamic>? mockItem;

  NewRepositoryOptions({
    required this.baseUrl,
    this.findUrl,
    this.addUrl,
    this.updateUrl,
    this.deleteUrl,
    this.mockItems,
    this.mockItem,
  });

  @override
  String getBaseUrl() {
    return this.baseUrl;
  }

  @override
  String? getFindUrl() {
    return this.findUrl;
  }

  @override
  String? getAddUrl() {
    return this.addUrl;
  }

  @override
  String? getUpdateUrl() {
    return this.updateUrl;
  }

  @override
  String? getDeleteUrl() {
    return this.deleteUrl;
  }

  @override
  List<Map<String, dynamic>>? getMockItems() {
    return this.mockItems;
  }

  @override
  Map<String, dynamic>? getMockItem() {
    return this.mockItem;
  }
}
